@@ -36,6 +36,12 @@ system_info()
 {
 	echo -e "\n###System Info###"
 
+	echo -e "#date"
+	date
+
+	echo -e "#ps axf"
+	ps axf
+
 	echo -e "#lsscsi"
 	lsscsi
 
@@ -72,8 +78,10 @@ adapter_info()
 		ethtool -i $DEVICE
 		echo -e "#ethtool offloads:"
 		ethtool -k $DEVICE
-		echo -e "#ifconfig:"
-		ifconfig $DEVICE
+		echo -e "#ip link:"
+		ip link show dev $DEVICE
+		echo -e "#ip addr:"
+		ip addr show dev $DEVICE
 	fi
 
 	echo -e "\n###Adapter INFO $PHYSDEV"
@@ -87,19 +95,15 @@ adapter_info()
 	ethtool -k $PHYSDEV
 	echo -e "#ethtool stats:"
 	ethtool -S $PHYSDEV
-	echo -e "#ifconfig:"
-	ifconfig $PHYSDEV
+	echo -e "#ip link:"
+	ip link show dev $PHYSDEV
+	echo -e "#ip addr:"
+	ip addr show dev $PHYSDEV
 }
 
-dcb_info()
+dcbtool_info()
 {
-	echo -e "\n###DCB INFO"
-	echo -e "#tc config"
-	tc qdisc
-	tc filter show dev $PHYSDEV | grep -v filter
-	echo -e "#service lldpad status:"
-	service lldpad status
-	echo -e "\n########## Showing dcb for $PHYSDEV"
+	echo -e "\n#### Showing dcbtool info for $PHYSDEV"
 	dcbtool -v
 	dcbtool gc $PHYSDEV dcb
 	echo -e "\n########## Getting dcb config for $PHYSDEV"
@@ -128,17 +132,79 @@ dcb_info()
 	dcbtool gp $PHYSDEV ll:0
 }
 
+lldptool_info()
+{
+    echo -e "\n#### Showing lldptool info for $PHYSDEV"
+    lldptool -v
+
+    echo -e "\n########## Showing last received TLV for $PHYSDEV"
+    lldptool -t -i $PHYSDEV -n
+
+    echo -e "\n########## Showing last sent TLV for $PHYSDEV"
+    lldptool -t -i $PHYSDEV
+
+    echo -e "\n########## Showing configured APP TLV for $PHYSDEV"
+    lldptool -t -i $PHYSDEV -V APP -c
+
+    echo -e "\n########## Showing configured PFC TLV for $PHYSDEV"
+    lldptool -t -i $PHYSDEV -V PFC -c
+
+    echo -e "\n########## Showing configured ETS-CFG TLV for $PHYSDEV"
+    lldptool -t -i $PHYSDEV -V ETS-CFG -c
+
+    echo -e "\n########## Showing configured ETS-REC TLV for $PHYSDEV"
+    lldptool -t -i $PHYSDEV -V ETS-REC -c
+}
+
+dcb_info()
+{
+	echo -e "\n###DCB INFO"
+	echo -e "#tc config"
+	tc qdisc
+	tc filter show dev $PHYSDEV | grep -v filter
+
+	## Intentionally allow both services status to show
+	## even though they should be mutually exclusive. If
+	## you see both in a dump you know there's a problem.
+	[ -f /etc/init.d/boot.lldpad ] &&
+	echo -e "#service boot.lldpad status:" &&
+	service boot.lldpad status
+	[ -f /etc/init.d/lldpad ] &&
+	echo -e "#service lldpad status:" &&
+	service lldpad status
+
+	which dcbtool 2>&1 > /dev/null
+	[ $? -eq 0 ] && dcbtool_info
+
+	which lldptool 2>&1 > /dev/null
+	[ $? -eq 0 ] && lldptool_info
+}
+
 fcoe_info()
 {
 	echo -e "\n###FCOE Info"
-	echo -e "#service fcoe status"
+
+	## Intentionally allow both services status to show
+	## even though they should be mutually exclusive. If
+	## you see both in a dump you know there's a problem.
+	[ -f /etc/init.d/boot.fcoe ] &&
+	echo -e "#service boot.fcoe status" &&
+	service boot.fcoe status
+	[ -f /etc/init.d/fcoe ] &&
+	echo -e "#service fcoe status" &&
 	service fcoe status
+
 	echo -e "#fcoeadm output "
 	fcoeadm -v
 	echo -e "#fcoeadm -i "
 	fcoeadm -i
 	echo -e "#fcoeadm -t "
 	fcoeadm -t
+
+	# Trigger fcoemon to dump its internal structures
+	# to the log file that we will capture later in
+	# this script.
+	kill -10 `pidof fcoemon`
 }
 
 bsg_info()
