@@ -60,7 +60,7 @@
 #include "fcoemon.h"
 #include "fcoe_clif.h"
 #include "fcoe_utils.h"
-#include "hbaapi.h"
+#include "sysfs_hba.h"
 #include "strarr.h"
 
 #include "fip.h"
@@ -169,7 +169,7 @@ static void fcm_dcbd_rx(void *);
 static void fcm_dcbd_event(char *, size_t);
 static void fcm_dcbd_cmd_resp(char *, cmd_status);
 static void fcm_netif_advance(struct fcm_netif *);
-static void fcm_fcoe_action(struct fcm_netif *, struct fcoe_port *);
+static void fcm_fcoe_action(struct fcoe_port *);
 static void fcp_set_next_action(struct fcoe_port *, enum fcp_action);
 static enum fcoe_status fcm_fcoe_if_action(char *, char *);
 
@@ -178,15 +178,15 @@ static enum fcoe_status fcm_fcoe_if_action(char *, char *);
  * "control" interfaces.
  */
 struct libfcoe_interface_template {
-	enum fcoe_status (*create)(struct fcm_netif *, struct fcoe_port *);
-	enum fcoe_status (*destroy)(struct fcm_netif *, struct fcoe_port *);
-	enum fcoe_status (*enable)(struct fcm_netif *, struct fcoe_port *);
-	enum fcoe_status (*disable)(struct fcm_netif *, struct fcoe_port *);
+	enum fcoe_status (*create)(struct fcoe_port *);
+	enum fcoe_status (*destroy)(struct fcoe_port *);
+	enum fcoe_status (*enable)(struct fcoe_port *);
+	enum fcoe_status (*disable)(struct fcoe_port *);
 };
 
 static const struct libfcoe_interface_template *libfcoe_control;
 
-static enum fcoe_status fcm_module_create(struct fcm_netif *ff, struct fcoe_port *p)
+static enum fcoe_status fcm_module_create(struct fcoe_port *p)
 {
 	enum fcoe_status rc;
 
@@ -216,17 +216,17 @@ static enum fcoe_status fcm_module_create(struct fcm_netif *ff, struct fcoe_port
 	return SUCCESS;
 }
 
-static enum fcoe_status fcm_module_destroy(struct fcm_netif *ff, struct fcoe_port *p)
+static enum fcoe_status fcm_module_destroy(struct fcoe_port *p)
 {
 	return fcm_fcoe_if_action(FCOE_DESTROY, p->ifname);
 }
 
-static enum fcoe_status fcm_module_enable(struct fcm_netif *ff, struct fcoe_port *p)
+static enum fcoe_status fcm_module_enable(struct fcoe_port *p)
 {
 	return fcm_fcoe_if_action(FCOE_ENABLE, p->ifname);
 }
 
-static enum fcoe_status fcm_module_disable(struct fcm_netif *ff, struct fcoe_port *p)
+static enum fcoe_status fcm_module_disable(struct fcoe_port *p)
 {
 	return fcm_fcoe_if_action(FCOE_DISABLE, p->ifname);
 }
@@ -238,13 +238,12 @@ static struct libfcoe_interface_template libfcoe_module_tmpl = {
 	.disable = fcm_module_disable,
 };
 
-static enum fcoe_status fcm_bus_enable(struct fcm_netif *ff,
-				       struct fcoe_port *p)
+static enum fcoe_status fcm_bus_enable(struct fcoe_port *p)
 {
 	return fcm_write_str_to_ctlr_attr(p->ctlr, FCOE_CTLR_ATTR_ENABLED, "1");
 }
 
-static int fcm_bus_configure(struct fcm_netif *ff, struct fcoe_port *p)
+static int fcm_bus_configure(struct fcoe_port *p)
 {
 	int rc;
 
@@ -255,8 +254,7 @@ static int fcm_bus_configure(struct fcm_netif *ff, struct fcoe_port *p)
 	return rc;
 }
 
-static enum fcoe_status fcm_bus_create(struct fcm_netif *ff,
-				       struct fcoe_port *p)
+static enum fcoe_status fcm_bus_create(struct fcoe_port *p)
 {
 	enum fcoe_status rc;
 
@@ -284,21 +282,19 @@ static enum fcoe_status fcm_bus_create(struct fcm_netif *ff,
 		return ENOSYSFS;
 	}
 
-	rc = fcm_bus_configure(ff, p);
+	rc = fcm_bus_configure(p);
 	if (!rc)
-		rc = fcm_bus_enable(ff, p);
+		rc = fcm_bus_enable(p);
 
 	return rc;
 }
 
-static enum fcoe_status fcm_bus_destroy(struct fcm_netif *ff,
-					struct fcoe_port *p)
+static enum fcoe_status fcm_bus_destroy(struct fcoe_port *p)
 {
 	return fcm_write_str_to_sysfs_file(FCOE_BUS_DESTROY, p->ifname);
 }
 
-static enum fcoe_status fcm_bus_disable(struct fcm_netif *ff,
-					struct fcoe_port *p)
+static enum fcoe_status fcm_bus_disable(struct fcoe_port *p)
 {
 	return fcm_write_str_to_ctlr_attr(p->ctlr, FCOE_CTLR_ATTR_ENABLED, "0");
 }
@@ -332,7 +328,6 @@ static int fcm_link_socket;
 static int fcm_link_seq;
 static void fcm_link_recv(void *);
 static void fcm_link_getlink(void);
-static int fcm_link_buf_check(size_t);
 static void clear_dcbd_info(struct fcm_netif *ff);
 static int fcoe_vid_from_ifname(const char *ifname);
 
@@ -340,9 +335,9 @@ static int fcoe_vid_from_ifname(const char *ifname);
  * Table for getopt_long(3).
  */
 static struct option fcm_options[] = {
-	{"debug", 0, NULL, 'd'},
+	{"debug", 1, NULL, 'd'},
 	{"legacy", 0, NULL, 'l'},
-	{"syslog", 0, NULL, 's'},
+	{"syslog", 1, NULL, 's'},
 	{"exec", 1, NULL, 'e'},
 	{"foreground", 0, NULL, 'f'},
 	{"version", 0, NULL, 'v'},
@@ -358,8 +353,7 @@ char progname[20];
  * large enough to fit and expand it if we ever do a read that almost fills it.
  */
 static char *fcm_link_buf;
-static size_t fcm_link_buf_size = 4096;	/* initial size */
-static const size_t fcm_link_buf_fuzz = 300;	/* "almost full" remainder */
+static size_t fcm_link_buf_size = 8192;	/* initial size */
 
 /*
  * A value must be surrounded by quates, e.g. "x".
@@ -524,6 +518,7 @@ static int fcm_read_config_files(void)
 		}
 		strncpy(file, CONFIG_DIR "/", sizeof(file));
 		strncat(file, dp->d_name, sizeof(file) - strlen(file));
+		file[sizeof(file) - 1] = '\0';
 		fp = fopen(file, "r");
 		if (!fp) {
 			FCM_LOG_ERR(errno, "Failed to read %s\n", file);
@@ -561,6 +556,9 @@ static int fcm_read_config_files(void)
 		if (!strncasecmp(val, "yes", 3) && rc == 1)
 			next->dcb_required = 1;
 
+		if (next->dcb_required == 1 && fcoe_config.dcb_init == 0)
+			fcoe_config.dcb_init = 1;
+
 		/* AUTO_VLAN */
 		rc = fcm_read_config_variable(file, val, sizeof(val),
 					      fp, CFG_IF_VAR_AUTOVLAN);
@@ -751,16 +749,18 @@ static void fcm_fc_event_handler(struct fc_nl_event *fc_event)
 	}
 }
 
-static void log_nlmsg_error(struct nlmsghdr *hp, int rlen, const char *str)
+static int log_nlmsg_error(struct nlmsghdr *hp, size_t rlen, const char *str)
 {
 	struct nlmsgerr *ep;
 
 	if (NLMSG_OK(hp, rlen)) {
-		ep = (struct nlmsgerr *)NLMSG_NEXT(hp, rlen);
+		ep = (struct nlmsgerr *)NLMSG_DATA(hp);
 		FCM_LOG_DBG("%s, err=%d, type=%d\n",
 			    str, ep->error, ep->msg.nlmsg_type);
+		return ep->error;
 	} else {
 		FCM_LOG("%s", str);
+		return 0;
 	}
 }
 
@@ -799,30 +799,31 @@ static void fcm_fc_event_log(struct fc_nl_event *fe)
 		{ FCH_EVT_LINK_UNKNOWN,         "link_unknown" },
 		{ FCH_EVT_VENDOR_UNIQUE,        "vendor_unique" },
 	};
-	int i;
+	unsigned int i;
 
 	for (i = 0; i < ARRAY_SIZE(fc_host_event_code_names); i++) {
 		if (fe->event_code == fc_host_event_code_names[i].value) {
 			/* only do u32 data even len is not, e.g. vendor */
-			FCM_LOG("FC_HOST_EVENT %d at %" PRIu64 " secs on host%d:"
-				"code %d=%s datalen %d data=%d\n",
-				fe->event_num, fe->seconds,
-				fe->host_no, fe->event_code,
-				fc_host_event_code_names[i].name,
-				fe->event_datalen, fe->event_data);
+			FCM_LOG_DBG("FC_HOST_EVENT %d at %" PRIu64 " secs on "
+				    "host%d code %d=%s datalen %d data=%d\n",
+				    fe->event_num, fe->seconds,
+				    fe->host_no, fe->event_code,
+				    fc_host_event_code_names[i].name,
+				    fe->event_datalen, fe->event_data);
 			break;
 		}
 	}
 
 }
 
-static void fcm_fc_event_recv(void *arg)
+static void fcm_fc_event_recv(UNUSED void *arg)
 {
 	struct nlmsghdr *hp;
 	struct fc_nl_event *fc_event;
-	int plen;
-	int rlen;
+	size_t plen;
+	size_t rlen;
 	char *buf;
+	int rc;
 
 	buf = malloc(DEF_RX_BUF_SIZE);
 
@@ -831,16 +832,17 @@ static void fcm_fc_event_recv(void *arg)
 		return;
 	}
 
-	rlen = read(fcm_fc_socket, buf, DEF_RX_BUF_SIZE);
-	if (!rlen)
+	rc = read(fcm_fc_socket, buf, DEF_RX_BUF_SIZE);
+	if (!rc)
 		goto free_buf;
 
-	if (rlen < 0) {
+	if (rc < 0) {
 		FCM_LOG_ERR(errno, "fc read error");
 		goto free_buf;
 	}
 
 	hp = (struct nlmsghdr *)buf;
+	rlen = rc;
 	for (hp = (struct nlmsghdr *)buf; NLMSG_OK(hp, rlen);
 	     hp = NLMSG_NEXT(hp, rlen)) {
 
@@ -855,7 +857,7 @@ static void fcm_fc_event_recv(void *arg)
 		plen = NLMSG_PAYLOAD(hp, 0);
 		fc_event = (struct fc_nl_event *)NLMSG_DATA(hp);
 		if (plen < sizeof(*fc_event)) {
-			FCM_LOG("too short (%d) to be an FC event", rlen);
+			FCM_LOG("too short (%zu) to be an FC event", rlen);
 			break;
 		}
 		fcm_fc_event_log(fc_event);
@@ -929,7 +931,7 @@ static int fcm_link_init(void)
 static struct fcoe_port *
 fcm_port_create(char *ifname, enum clif_flags flags, int cmd);
 
-struct fcoe_port *fcm_new_vlan(int ifindex, int vid, bool vn2vn)
+static struct fcoe_port *fcm_new_vlan(int ifindex, int vid, bool vn2vn)
 {
 	char real_name[IFNAMSIZ];
 	char vlan_name[IFNAMSIZ];
@@ -938,6 +940,7 @@ struct fcoe_port *fcm_new_vlan(int ifindex, int vid, bool vn2vn)
 		[false] = CLIF_FLAGS_FABRIC,
 		[true] = CLIF_FLAGS_VN2VN,
 	};
+	int rc;
 
 	if (vn2vn)
 		FCM_LOG_DBG("Auto VLAN found vn2vn on VID %d\n", vid);
@@ -946,8 +949,15 @@ struct fcoe_port *fcm_new_vlan(int ifindex, int vid, bool vn2vn)
 
 	if (rtnl_find_vlan(ifindex, vid, vlan_name)) {
 		rtnl_get_linkname(ifindex, real_name);
-		snprintf(vlan_name, sizeof(vlan_name), FCOE_VLAN_FORMAT,
-			 real_name, vid);
+		rc = snprintf(vlan_name, sizeof(vlan_name), FCOE_VLAN_FORMAT,
+			      real_name, vid);
+		if (rc < 0 || (size_t) rc >= sizeof(vlan_name)) {
+			FCM_LOG("Warning: Generating FCoE VLAN device name for"
+				"interface %s VLAN %d: format resulted in a"
+				"name larger than IFNAMSIZ\n", real_name, vid);
+			vlan_name[sizeof(vlan_name) - 1] = 0;
+			FCM_LOG("\tTruncating VLAN name to %s\n", vlan_name);
+		}
 		vlan_create(ifindex, vid, vlan_name);
 	}
 	rtnl_set_iff_up(0, vlan_name);
@@ -959,8 +969,8 @@ struct fcoe_port *fcm_new_vlan(int ifindex, int vid, bool vn2vn)
 	return p;
 }
 
-
-int fcm_vlan_disc_handler(struct fiphdr *fh, struct sockaddr_ll *sa, void *arg)
+static int
+fcm_vlan_disc_handler(struct fiphdr *fh, struct sockaddr_ll *sa, void *arg)
 {
 	int vid;
 	unsigned char mac[ETHER_ADDR_LEN];
@@ -1054,7 +1064,7 @@ static int fcm_vlan_disc_socket(struct fcoe_port *p)
 	int fd;
 	int origdev = 1;
 
-	fd = fip_socket(p->ifindex, FIP_NONE);
+	fd = fip_socket(p->ifindex, p->mac, FIP_NONE);
 	if (fd < 0) {
 		FCM_LOG_ERR(errno, "socket error");
 		return fd;
@@ -1076,6 +1086,7 @@ static void fcm_vlan_dev_real_dev(char *vlan_ifname, char *real_ifname)
 {
 	int fd;
 	struct vlan_ioctl_args ifv;
+	int rc;
 
 	real_ifname[0] = '\0';
 
@@ -1088,9 +1099,23 @@ static void fcm_vlan_dev_real_dev(char *vlan_ifname, char *real_ifname)
 
 	memset(&ifv, 0, sizeof(ifv));
 	ifv.cmd = GET_VLAN_REALDEV_NAME_CMD;
-	strncpy(ifv.device1, vlan_ifname, strlen(vlan_ifname)+1);
-	if (ioctl(fd, SIOCGIFVLAN, &ifv) == 0)
-		strncpy(real_ifname, ifv.u.device2, strlen(ifv.u.device2)+1);
+	if (strlen(vlan_ifname) > (sizeof(ifv.device1) - 1)) {
+		FCM_LOG_ERR(ENOSPC, "no room for vlan ifname");
+		goto close_fd;
+	}
+
+	rc = snprintf(ifv.device1, IFNAMSIZ, "%s", vlan_ifname);
+	if (rc < 0 || rc >= IFNAMSIZ)
+		goto close_fd;
+
+	if (ioctl(fd, SIOCGIFVLAN, &ifv) == 0) {
+		rc = snprintf(real_ifname, IFNAMSIZ, "%s", ifv.u.device2);
+		if (rc < 0 || rc >= IFNAMSIZ) {
+			real_ifname[0] = '\0';
+			goto close_fd;
+		}
+	}
+close_fd:
 	close(fd);
 }
 
@@ -1100,7 +1125,7 @@ static void fcm_vlan_dev_real_dev(char *vlan_ifname, char *real_ifname)
  * This function parses the linkinfo rtattr and returns
  * 1 if it is kind vlan otherwise returns 0.
  */
-int fcm_is_linkinfo_vlan(struct rtattr *ap)
+static int fcm_is_linkinfo_vlan(struct rtattr *ap)
 {
 	struct rtattr *info;
 	int len;
@@ -1343,6 +1368,7 @@ STR_ARR(ieee_states, "Unknown", "Out of range",
 	[IEEE_INIT] = "IEEE_INIT",
 	[IEEE_GET_STATE] = "IEEE_GET_STATE",
 	[IEEE_DONE] = "IEEE_DONE",
+	[IEEE_ACTIVE] = "IEEE_ACTIVE",
 );
 
 static void
@@ -1359,12 +1385,8 @@ ieee_state_set(struct fcm_netif *ff, enum ieee_state new_state)
 				getstr(&ieee_states, new_state));
 	}
 
-	if (new_state == IEEE_GET_STATE) {
-		ff->ieee_state = new_state;
+	if (new_state == IEEE_GET_STATE)
 		clear_ieee_info(ff);
-		ieee_get_req(ff);
-		return;
-	}
 
 	ff->ieee_state = new_state;
 	ff->ieee_resp_pending = 0;
@@ -1413,8 +1435,7 @@ static void fcm_dcbd_state_set(struct fcm_netif *ff,
 	ff->response_pending = 0;
 }
 
-static int fip_recv_vlan_req(struct fiphdr *fh, struct sockaddr_ll *ssa,
-			     struct fcoe_port *sp)
+static int fip_recv_vlan_req(struct sockaddr_ll *ssa, struct fcoe_port *sp)
 {
 	struct fip_tlv_vlan *vlan_tlvs = NULL;
 	int vlan_count = 0;
@@ -1473,7 +1494,7 @@ static int fip_vlan_disc_handler(struct fiphdr *fh, struct sockaddr_ll *sa,
 	switch (fh->fip_subcode) {
 	case FIP_VLAN_REQ:
 		FCM_LOG_DBG("received VLAN req, subcode=%d\n", fh->fip_subcode);
-		rc = fip_recv_vlan_req(fh, sa, p);
+		rc = fip_recv_vlan_req(sa, p);
 		break;
 	default:
 		FCM_LOG_DBG("ignored FIP VLAN packet with subcode %d\n",
@@ -1497,7 +1518,7 @@ static void init_fip_vn2vn_responder(struct fcoe_port *p)
 	if (p->fip_responder_socket >= 0)
 		return;
 
-	s = fip_socket(p->ifindex, FIP_ALL_VN2VN);
+	s = fip_socket(p->ifindex, p->mac, FIP_ALL_VN2VN);
 	if (s < 0) {
 		FCM_LOG_ERR(errno, "%s: Failed to get fip socket\n", p->ifname);
 		return;
@@ -1582,7 +1603,7 @@ static int fcoe_vid_from_ifname(const char *ifname)
 	return -1;
 }
 
-void fcm_process_link_msg(struct ifinfomsg *ip, int len, unsigned type)
+static void fcm_process_link_msg(struct ifinfomsg *ip, int len, unsigned type)
 {
 	struct fcoe_port *p;
 	struct rtattr *ap;
@@ -1645,8 +1666,10 @@ void fcm_process_link_msg(struct ifinfomsg *ip, int len, unsigned type)
 		/* try to find the real device name */
 		real_dev[0] = '\0';
 		fcm_vlan_dev_real_dev(ifname, real_dev);
-		if (strlen(real_dev))
-			strncpy(p->real_ifname, real_dev, strlen(real_dev)+1);
+		if (strlen(real_dev)) {
+			strncpy(p->real_ifname, real_dev, IFNAMSIZ);
+			p->real_ifname[IFNAMSIZ - 1] = '\0';
+		}
 		if (p->ready)
 			update_fcoe_port_state(p, type, operstate,
 					       FCP_CFG_IFNAME);
@@ -1658,7 +1681,8 @@ void fcm_process_link_msg(struct ifinfomsg *ip, int len, unsigned type)
 		if (p) {
 			p->ifindex = ifindex;
 			memcpy(p->mac, mac, ETHER_ADDR_LEN);
-			strncpy(p->real_ifname, ifname, strlen(ifname)+1);
+			strncpy(p->real_ifname, ifname, IFNAMSIZ);
+			p->real_ifname[IFNAMSIZ - 1] = '\0';
 			update_fcoe_port_state(p, type, operstate,
 					       FCP_REAL_IFNAME);
 		}
@@ -1786,7 +1810,9 @@ static void fcm_process_ieee_msg(struct nlmsghdr *nlh)
 	if (rta_parent->rta_type != DCB_ATTR_IFNAME)
 		return;
 
-	strncpy(ifname, NLA_DATA(rta_parent), sizeof(ifname));
+	strncpy(ifname, NLA_DATA(rta_parent), IFNAMSIZ);
+	ifname[IFNAMSIZ - 1] = '\0';
+
 	ff = fcm_netif_lookup_create(ifname);
 	if (!ff) {
 		FCM_LOG("Processing IEEE message: %s not found or created\n",
@@ -1841,18 +1867,32 @@ static void fcm_process_ieee_msg(struct nlmsghdr *nlh)
 		ieee_state_set(ff, IEEE_DONE);
 }
 
-static void fcm_link_recv(void *arg)
+static void fcm_link_recv(UNUSED void *arg)
 {
 	int rc;
 	char *buf;
 	struct nlmsghdr *hp;
 	struct ifinfomsg *ip;
 	unsigned type;
-	int plen;
-	int rlen;
+	size_t plen;
+	size_t rlen;
+
+	/* check to make sure our receive buffer is large enough,
+	 * or scale it up as needed */
+	rc = recv(fcm_link_socket, NULL, 0, MSG_PEEK | MSG_TRUNC);
+	if (rc > (int)fcm_link_buf_size) {
+		FCM_LOG_DBG("resizing link buf to %d bytes\n", rc);
+		void *resize = realloc(fcm_link_buf, rc);
+		if (resize) {
+			fcm_link_buf = resize;
+			fcm_link_buf_size = rc;
+		} else {
+			FCM_LOG_ERR(errno, "Failed to allocate link buffer");
+		}
+	}
 
 	buf = fcm_link_buf;
-	rc = read(fcm_link_socket, buf, fcm_link_buf_size);
+	rc = recv(fcm_link_socket, buf, fcm_link_buf_size, 0);
 	if (rc <= 0) {
 		if (rc < 0)
 			FCM_LOG_ERR(errno, "Error reading from "
@@ -1861,11 +1901,6 @@ static void fcm_link_recv(void *arg)
 		return;
 	}
 
-	if (fcm_link_buf_check(rc)) {
-		fcm_link_getlink();
-		return;
-	}
-
 	hp = (struct nlmsghdr *)buf;
 	rlen = rc;
 	for (hp = (struct nlmsghdr *)buf; NLMSG_OK(hp, rlen);
@@ -1876,7 +1911,7 @@ static void fcm_link_recv(void *arg)
 			break;
 
 		if (hp->nlmsg_type == NLMSG_ERROR) {
-			log_nlmsg_error(hp, rlen, "nlmsg error");
+			rc = log_nlmsg_error(hp, rlen, "nlmsg error");
 			break;
 		}
 
@@ -1907,6 +1942,11 @@ static void fcm_link_recv(void *arg)
 			break;
 		}
 	}
+	if (rc == -EBUSY) {
+		FCM_LOG_DBG("%s: netlink returned -EBUSY, retry\n",
+			      __func__);
+		fcm_link_getlink();
+	}
 }
 
 /*
@@ -1930,34 +1970,9 @@ static void fcm_link_getlink(void)
 	msg.nl.nlmsg_pid = getpid();
 	msg.ifi.ifi_family = AF_UNSPEC;
 	msg.ifi.ifi_type = ARPHRD_ETHER;
-	rc = write(fcm_link_socket, &msg, sizeof(msg));
+	rc = send(fcm_link_socket, &msg, sizeof(msg), 0);
 	if (rc < 0)
-		FCM_LOG_ERR(errno, "write error");
-}
-
-/*
- * Check for whether buffer needs to grow based on amount read.
- * Free's the old buffer so don't use that after this returns non-zero.
- */
-static int fcm_link_buf_check(size_t read_len)
-{
-	char *buf;
-	size_t len = read_len;
-
-	if (len > fcm_link_buf_size - fcm_link_buf_fuzz) {
-		len = fcm_link_buf_size;
-		len = len + len / 2;	/* grow by 50% */
-		buf = malloc(len);
-		if (buf != NULL) {
-			free(fcm_link_buf);
-			fcm_link_buf = buf;
-			fcm_link_buf_size = len;
-			return 1;
-		} else {
-			FCM_LOG_ERR(errno, "failed to allocate link buffer");
-		}
-	}
-	return 0;
+		FCM_LOG_ERR(errno, "send error");
 }
 
 static void fcm_fcoe_init(void)
@@ -2093,7 +2108,7 @@ static int fcm_dcbd_connect(void)
 	return 1;
 }
 
-static void fcm_dcbd_timeout(void *arg)
+static void fcm_dcbd_timeout(UNUSED void *arg)
 {
 	if (fcm_clif->cl_ping_pending > 0) {
 		fcm_dcbd_request("D");	/* DETACH_CMD */
@@ -2202,11 +2217,12 @@ static void fcm_dcbd_rx(void *arg)
 	rc = read(clif->cl_fd, buf, sizeof(buf) - 1);
 	if (rc < 0)
 		FCM_LOG_ERR(errno, "read");
-	else if ((rc > 0) && (rc < sizeof(buf))) {
+	else if (rc > 0 && rc < (int)sizeof(buf)) {
 		buf[rc] = '\0';
 		len = strlen(buf);
 		ASSERT(len <= rc);
 
+		FCM_LOG_DBG("recv '%s', len=%d bytes succeeded", buf, (int)len);
 		switch (buf[CLIF_RSP_MSG_OFF]) {
 		case CMD_RESPONSE:
 			st = fcm_get_hex(buf + CLIF_STAT_OFF, CLIF_STAT_LEN,
@@ -2640,7 +2656,7 @@ static void fcm_dcbd_get_oper(struct fcm_netif *ff, char *resp, char *cp)
 
 	if (ep) {
 		FCM_LOG_DEV(ff, "Invalid get oper response "
-			    "parse error byte %ld, resp %s", ep - cp, cp);
+			    "parse error byte %td, resp %s", ep - cp, cp);
 		fcm_dcbd_state_set(ff, FCD_ERROR);
 	} else {
 		if (val && fcoe_config.debug)
@@ -2908,20 +2924,40 @@ static void fcm_send_fip_request(const struct fcoe_port *p)
 void fcm_vlan_disc_timeout(void *arg)
 {
 	struct fcoe_port *p = arg;
+	int s;
+
 	FCM_LOG_DBG("%s: VLAN discovery TIMEOUT [%d]",
 		    p->ifname, p->vlan_disc_count);
 	p->vlan_disc_count++;
+	if (p->fip_socket < 0) {
+		s = fcm_vlan_disc_socket(p);
+		if (s < 0) {
+			FCM_LOG_ERR(errno, "Could not acquire fip socket.\n");
+			goto set_timeout;
+		}
+		p->fip_socket = s;
+		p->vlan_disc_count = 1;
+	}
 	fcm_send_fip_request(p);
+set_timeout:
 	sa_timer_set(&p->vlan_disc_timer, FCM_VLAN_DISC_TIMEOUT);
 }
 
-int fcm_start_vlan_disc(struct fcoe_port *p)
+static int fcm_start_vlan_disc(struct fcoe_port *p)
 {
 	int s;
 	if (p->fip_socket < 0) {
 		s = fcm_vlan_disc_socket(p);
-		if (s < 0)
+		if (s < 0) {
+			/*
+			 * If we can't open the socket set the timeout
+			 * anyways so we will retry sending the fipvlan
+			 * request.
+			 */
+			FCM_LOG_ERR(errno, "Failed to open socket, setting VLAN DISC timer.\n");
+			sa_timer_set(&p->vlan_disc_timer, FCM_VLAN_DISC_TIMEOUT);
 			return s;
+		}
 		p->fip_socket = s;
 	}
 	p->vlan_disc_count = 1;
@@ -2936,7 +2972,7 @@ int fcm_start_vlan_disc(struct fcoe_port *p)
  *         action = 2      Create the FCoE interface
  *         action = 3      Reset the interface
  */
-static void fcm_fcoe_action(struct fcm_netif *ff, struct fcoe_port *p)
+static void fcm_fcoe_action(struct fcoe_port *p)
 {
 	struct fcoe_port *vp;
 	char path[MAX_PATH_LEN];
@@ -2945,7 +2981,7 @@ static void fcm_fcoe_action(struct fcm_netif *ff, struct fcoe_port *p)
 	switch (p->action) {
 	case FCP_CREATE_IF:
 		FCM_LOG_DBG("OP: CREATE %s\n", p->ifname);
-		rc = libfcoe_control->create(ff, p);
+		rc = libfcoe_control->create(p);
 		if (rc) {
 			FCM_LOG_DBG("Failed to create FCoE interface "
 				    "for %s, rc is %d\n", p->ifname, rc);
@@ -2971,12 +3007,12 @@ static void fcm_fcoe_action(struct fcm_netif *ff, struct fcoe_port *p)
 			rc = SUCCESS;
 			break;
 		}
-		rc = libfcoe_control->destroy(ff, p);
+		rc = libfcoe_control->destroy(p);
 		p->fchost[0] = '\0';
 		break;
 	case FCP_ENABLE_IF:
 		FCM_LOG_DBG("OP: ENABLE %s\n", p->ifname);
-		rc = libfcoe_control->enable(ff, p);
+		rc = libfcoe_control->enable(p);
 		break;
 	case FCP_DISABLE_IF:
 		FCM_LOG_DBG("OP: DISABLE %s\n", p->ifname);
@@ -2992,7 +3028,7 @@ static void fcm_fcoe_action(struct fcm_netif *ff, struct fcoe_port *p)
 			}
 			break;
 		}
-		rc = libfcoe_control->disable(ff, p);
+		rc = libfcoe_control->disable(p);
 		break;
 	case FCP_RESET_IF:
 		FCM_LOG_DBG("OP: RESET %s\n", p->ifname);
@@ -3036,6 +3072,44 @@ static void fcm_fcoe_action(struct fcm_netif *ff, struct fcoe_port *p)
 	p->last_action = p->action;
 }
 
+/*
+ * Called for all ports.  For FCoE ports and candidates,
+ * get IEEE DCBX information and set the next action.
+ */
+static void fcm_netif_ieee_advance(struct fcm_netif *ff)
+{
+	enum fcp_action action;
+
+	ASSERT(ff);
+	ASSERT(fcm_clif);
+
+	if (fcm_clif->cl_busy)
+		return;
+
+	if (ff->ieee_resp_pending)
+		return;
+
+	switch (ff->ieee_state) {
+	case IEEE_INIT:
+		break;
+	case IEEE_GET_STATE:
+		ieee_get_req(ff);
+		break;
+	case IEEE_DONE:
+		action = validate_ieee_info(ff);
+		if (action == FCP_ACTIVATE_IF) {
+			fcp_action_set(ff->ifname, action);
+			ieee_state_set(ff, IEEE_ACTIVE);
+		}
+		break;
+	case IEEE_ACTIVE:
+		/* TBD enable and disable if needed in IEEE mode */
+		break;
+	default:
+		break;
+	}
+}
+
 /*
  * Called for all ports.  For FCoE ports and candidates,
  * get information and send to dcbd.
@@ -3156,7 +3230,7 @@ static void fcm_netif_advance(struct fcm_netif *ff)
  *       as necessary with a CREATE or DESTROY next action.
  * 2.  Process FCoE port list - handle next actions, update states, clean up
  */
-static void fcm_handle_changes()
+static void fcm_handle_changes(void)
 {
 	struct fcm_netif *ff;
 	struct fcoe_port *p;
@@ -3164,8 +3238,10 @@ static void fcm_handle_changes()
 	/*
 	 * Perform pending actions (dcbd queries) on network interfaces.
 	 */
-	TAILQ_FOREACH(ff, &fcm_netif_head, ff_list)
+	TAILQ_FOREACH(ff, &fcm_netif_head, ff_list) {
 		fcm_netif_advance(ff);
+		fcm_netif_ieee_advance(ff);
+	}
 
 	/*
 	 * Perform actions on FCoE ports
@@ -3183,7 +3259,7 @@ static void fcm_handle_changes()
 			goto next_port;
 		}
 
-		fcm_fcoe_action(ff, p);
+		fcm_fcoe_action(p);
 
 		fcp_set_next_action(p, FCP_WAIT);
 next_port:
@@ -3195,9 +3271,9 @@ static void fcm_usage(void)
 {
 	printf("Usage: %s\n"
 	       "\t [-f|--foreground]\n"
-	       "\t [-d|--debug]\n"
+	       "\t [-d|--debug=yes|no]\n"
 	       "\t [-l|--legacy]\n"
-	       "\t [-s|--syslog]\n"
+	       "\t [-s|--syslog=yes|no]\n"
 	       "\t [-v|--version]\n"
 	       "\t [-h|--help]\n\n", progname);
 	exit(1);
@@ -3460,9 +3536,10 @@ static void fcm_srv_receive(void *arg)
 	socklen_t fromlen = sizeof(struct sockaddr_un);
 	struct sock_info *reply = NULL;
 	char buf[MAX_MSGBUF], rbuf[MAX_MSGBUF];
-	char ifname[sizeof(data->ifname) + 1];
+	char ifname[IFNAMSIZ + 1];
 	enum fcoe_status rc = EFAIL;
 	int res, cmd, snum;
+	size_t size;
 
 	snum = srv_info->srv_sock;
 	res = recvfrom(snum, buf, sizeof(buf) - 1,
@@ -3474,22 +3551,23 @@ static void fcm_srv_receive(void *arg)
 	}
 
 	data = (struct clif_data *)buf;
-	if (res < sizeof(*data)) {
-		if (res < sizeof(*data) - sizeof(data->flags)) {
+	size = res;
+	if (size < sizeof(*data)) {
+		if (size < sizeof(*data) - sizeof(data->flags)) {
 			FCM_LOG_ERR(EMSGSIZE,
 				    "Message too short from socket %d", snum);
 			rc = EBADCLIFMSG;
 			goto err;
 		}
 		data->flags = 0;
-	} else if (res > sizeof(*data)) {
+	} else if (size > sizeof(*data)) {
 		FCM_LOG_ERR(EMSGSIZE, "Message too long from socket %d", snum);
 		rc = EBADCLIFMSG;
 		goto err;
 	}
 
 	cmd = data->cmd;
-	strncpy(ifname, data->ifname, sizeof(data->ifname));
+	strncpy(ifname, data->ifname, IFNAMSIZ);
 	ifname[sizeof(data->ifname)] = 0;
 
 	if (cmd != CLIF_PID_CMD) {
@@ -3547,12 +3625,54 @@ err:
 	sendto(snum, rbuf, MSG_RBUF, 0, (struct sockaddr *)&from, fromlen);
 }
 
+static int fcm_systemd_socket(void)
+{
+	char *env, *ptr;
+	unsigned int p, l;
+
+	env = getenv("LISTEN_PID");
+	if (!env)
+		return -1;
+
+	p = strtoul(env, &ptr, 10);
+	if (ptr && ptr == env) {
+		FCM_LOG_DBG("Invalid value '%s' for LISTEN_PID\n", env);
+		return -1;
+	}
+	if ((pid_t)p != getpid()) {
+		FCM_LOG_DBG("Invalid PID '%d' from LISTEN_PID\n", p);
+		return -1;
+	}
+	env = getenv("LISTEN_FDS");
+	if (!env) {
+		FCM_LOG_DBG("LISTEN_FDS is not set\n");
+		return -1;
+	}
+	l = strtoul(env, &ptr, 10);
+	if (ptr && ptr == env) {
+		FCM_LOG_DBG("Invalid value '%s' for LISTEN_FDS\n", env);
+		return -1;
+	}
+	if (l != 1) {
+		FCM_LOG_DBG("LISTEN_FDS specified %d fds\n", l);
+		return -1;
+	}
+	/* systemd returns fds with an offset of '3' */
+	return 3;
+}
+
 static int fcm_srv_create(struct fcm_srv_info *srv_info)
 {
 	socklen_t addrlen;
 	struct sockaddr_un addr;
 	int rc = 0;
 
+	srv_info->srv_sock = fcm_systemd_socket();
+	if (srv_info->srv_sock > 0) {
+		FCM_LOG_DBG("Using systemd socket\n");
+		goto out_done;
+	}
+
 	srv_info->srv_sock = socket(AF_LOCAL, SOCK_DGRAM, 0);
 	if (srv_info->srv_sock < 0) {
 		FCM_LOG_ERR(errno, "Failed to create socket\n");
@@ -3571,7 +3691,7 @@ static int fcm_srv_create(struct fcm_srv_info *srv_info)
 		rc = errno;
 		goto err_close;
 	}
-
+out_done:
 	sa_select_add_fd(srv_info->srv_sock, fcm_srv_receive,
 			 NULL, NULL, srv_info);
 
@@ -3603,26 +3723,34 @@ int main(int argc, char **argv)
 	memset(&fcoe_config, 0, sizeof(fcoe_config));
 
 	strncpy(progname, basename(argv[0]), sizeof(progname));
+	progname[sizeof(progname) - 1] = '\0';
+
 	sa_log_prefix = progname;
 	sa_log_flags = 0;
 	openlog(sa_log_prefix, LOG_CONS, LOG_DAEMON);
 
-	while ((c = getopt_long(argc, argv, "fdhlsv",
+	while ((c = getopt_long(argc, argv, "fd:hls:v",
 				fcm_options, NULL)) != -1) {
 		switch (c) {
 		case 'f':
 			fcm_fg = 1;
 			break;
 		case 'd':
-			fcoe_config.debug = 1;
-			enable_debug_log(1);
+			if (!strncmp(optarg, "yes", 3) ||
+			    !strncmp(optarg, "YES", 3)) {
+				fcoe_config.debug = 1;
+				enable_debug_log(1);
+			}
 			break;
 		case 'l':
 			force_legacy = true;
 			break;
 		case 's':
-			fcoe_config.use_syslog = 1;
-			enable_syslog(1);
+			if (!strncmp(optarg, "yes", 3) ||
+			    !strncmp(optarg, "YES", 3)) {
+				fcoe_config.use_syslog = 1;
+				enable_syslog(1);
+			}
 			break;
 		case 'v':
 			printf("%s\n", FCOE_UTILS_VERSION);
@@ -3636,11 +3764,6 @@ int main(int argc, char **argv)
 	if (argc != optind)
 		fcm_usage();
 
-	if (!fcm_fg && daemon(0, !fcoe_config.use_syslog)) {
-		FCM_LOG("Starting daemon failed");
-		exit(EXIT_FAILURE);
-	}
-
 	umask(0);
 
 	/*
@@ -3687,30 +3810,54 @@ int main(int argc, char **argv)
 	}
 
 	fcm_fcoe_init();
-	fcm_fc_events_init();
-	fcm_link_init();	/* NETLINK_ROUTE protocol */
-	fcm_dcbd_init();
-	fcm_srv_create(&srv_info);
+	rc = fcm_fc_events_init();
+	if (rc != 0)
+		exit(1);
+
+	rc = fcm_link_init();	/* NETLINK_ROUTE protocol */
+	if (rc != 0)
+		goto err_cleanup;
+
+	if (fcoe_config.dcb_init)
+		fcm_dcbd_init();
+
+	rc = fcm_srv_create(&srv_info);
+	if (rc == EADDRINUSE) {
+		FCM_LOG("Daemon already running OR Failed to bind socket so exiting!\n");
+		exit(1);
+	}
+	if (rc != 0)
+		goto err_cleanup;
+
+	if (!fcm_fg && daemon(0, !fcoe_config.use_syslog)) {
+		FCM_LOG("Starting daemon failed");
+		goto err_cleanup;
+	}
+
 	sa_select_set_callback(fcm_handle_changes);
 
 	rc = sa_select_loop();
 	if (rc < 0) {
 		FCM_LOG_ERR(rc, "select error\n");
-		exit(EXIT_FAILURE);
+		goto err_cleanup;
 	}
 	fcm_dcbd_shutdown();
 	fcm_srv_destroy(&srv_info);
 	if (rc == SIGHUP)
 		fcm_cleanup();
 	return 0;
+
+err_cleanup:
+	fcm_cleanup();
+	exit(1);
 }
 
 /*******************************************************
  *         The following are debug routines            *
  *******************************************************/
-static void add_msg_to_buf(char *buf, int maxlen, char *msg, char *prefix)
+static void add_msg_to_buf(char *buf, size_t maxlen, char *msg, char *prefix)
 {
-	int len = strlen(buf);
+	size_t len = strlen(buf);
 
 	if (len + strlen(msg) + strlen(prefix) < maxlen)
 		sprintf(buf+len, "%s%s", prefix, msg);
